#include "workspace.h"
#include "obstacle.h"
#include "util.h"
#include <assert.h>

Obstacle::Obstacle(istream& obstData, Workspace *myWork):work(myWork)
{
  readObstData(obstData);
  buildQuinRep();
}

const bool Obstacle::pointInside(const Point2& p) const {
  int numCross = 0;

  for (int i = 0; i < nvert; i++) {
    const Point2& start = work->getPointRef(vertices[i]);
    const Point2& end = work->getPointRef(vertices[(i + 1) % nvert]);
    if (((start.y > p.y) && (end.y <= p.y)) ||
	((end.y > p.y) && (start.y <= p.y))) {
      if ((((end.x-p.x)*(start.y-p.y))-((start.x-p.x)*(end.y-p.y)))/
	  (start.y-end.y) > 0)
	numCross++;
    }
  }

  return ((numCross % 2) == 1);
}

// Private member functions

void Obstacle::readObstData(istream& obstData)
{
  if (verbose_input)
    cout << "Adding polygon: ";
  obstData >> nvert;
  if (verbose_input)
    cout << "n = " << nvert << " Vertices: ";
  vertices = new int[nvert];
  assert(vertices);
  for (int i = 0; i < nvert; i++) {
    obstData >> vertices[i];
    if (verbose_input)
      cout << vertices[i] << " ";
  }
  if (verbose_input)
    cout << endl;

  maxpen = 0;
  for (int i = 0; i < nvert; i++)
    for (int j = i + 1; j < nvert; j++) {
      double dist = pabs2(psub2(work->getPointRef(vertices[i]), 
				work->getPointRef(vertices[j])));
      if ((maxpen == 0) || (-dist > maxpen))
	maxpen = -dist;
    }

  maxpen = -.1;

  cout << "MAXPEN = " << maxpen << endl;
}

void Obstacle::buildQuinRep()
{
  bgnsobj2();
  for (int i = 0; i < nvert; i++) {
    Point2 p[2];
    p[0] = work->getPointRef(vertices[i%nvert]);
    p[1] = work->getPointRef(vertices[(i+1)%nvert]);
    sline2(p, 0, 1, NULL);
  }
  quinRep = endsobj2();
}





