/* pathedge.h
   Stephen Sorkib
   This file defines the PathEdge class.
   */

#ifndef _pathedge_h
#define _pathedge_h

#include "darray.h"

class PathNode;
class PathPlanner;
class Workspace;


struct MiniEdge {
  //PathNode *from, *to;
  double penFrac;
};

class PathEdge {
 public:
  static bool makeEdgePtr(PathNode *start, PathNode *end, double penFrac,
			  PathPlanner *pathPlanner, 
			  const Workspace *workspace);
  static PathEdge *makeEdgePtr(PathNode *start, PathNode *end, double penFrac,
			       DArray<MiniEdge>& subEdges, 
			       PathPlanner *pathPlanner, 
			       const Workspace *workspace);


  void moveToFree(double newPenFrac);
  bool freePath(double maxPenFrac);
  void moveNode() {penFrac = 100;}

  void removeNode(const PathNode *node);
  ~PathEdge();
  PathNode *getNode(int nodeNum);

  int numSegments() const;
  friend ostream& operator<<(ostream& out, const PathEdge& pathEdge); 

  static void saveEdgesVect(DArray<PathEdge *>& edges, const char *outFileName);
  static bool connected(PathNode* from, PathNode* to, DArray<PathEdge *>& edges, DArray<PathNode *>& nodes, bool destructive);

 private:
  PathNode *from, *to;
  double penFrac;
  DArray<PathNode *> intermediates;
  PathPlanner *pathPlanner;
  const Workspace *workspace;
  DArray<MiniEdge> subEdges;

  PathEdge(PathNode *start, PathNode *end, double myPenFrac, 
	   PathPlanner *myPlanner, const Workspace *myWorkspace);
  
  static bool makePath(PathNode *start, PathNode *end, double maxPenFrac, 
		       PathPlanner *myPlanner, const Workspace *workspace, 
		       int depthLeft, DArray<MiniEdge>& subEdges);


  static bool connectedHelp(PathNode* from, PathNode* to, DArray<PathEdge *>& edges, DArray<PathNode *>& visited);
};

#endif
